from findiff import FinDiff
from plt_grids_with_stencils import *

laplace = FinDiff(0, 1, 2) + FinDiff(1, 1, 2)

plt.rcParams.update({'font.size': 20, "mathtext.fontset": "cm"})
stencil_set = laplace.stencil((5, 5))

stencil = stencil_set.data[('C', 'C')]
fig, ax = plt.subplots(figsize=(8, 8))
plot_axes(ax, stencil=list(stencil.keys()), grid_kernel=[0], coeffs=list(stencil.values()), markersize=40)
fig.savefig("stencil_laplace2d_center.png")

stencil = stencil_set.data[('L', 'C')]
fig, ax = plt.subplots(figsize=(8, 8))
plot_axes(ax, stencil=list(stencil.keys()), grid_kernel=[0], coeffs=list(stencil.values()), markersize=40)
fig.savefig("stencil_laplace2d_border.png")

stencil = stencil_set.data[('L', 'L')]
fig, ax = plt.subplots(figsize=(8, 8))
plot_axes(ax, stencil=list(stencil.keys()), grid_kernel=[0], coeffs=list(stencil.values()), markersize=40)
fig.savefig("stencil_laplace2d_corner.png")

print(stencil_set)



import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.patches import FancyArrowPatch


import numpy as np

plt.rcParams.update({"font.size": 18, "font.family":"serif",
                     "mathtext.fontset": "dejavuserif"})

fig, ax = plt.subplots(layout="constrained")


xr = (0.7, 2)

# The continuous function
x_fine = np.linspace(*xr, 100)
f_fine = np.sin(x_fine)
ax.plot(x_fine, f_fine, "-")

# The function on the grid
x = np.linspace(*xr, 7)
f = np.sin(x)
ax.plot(x, f, 'o')


ax.set_xlabel("")
ax.set_ylabel("$f\,(x)$")

labels = [r"$x_{i%+d}$" % (i-3) for i in range(len(x))]
labels[3] = r"$x_i$"

ax.set_xticks(ticks=x, labels=labels)
ax.set_yticks(ticks=[], labels=[])

ax.set_ylim(0.6, 1.02)

for i, (xi, fi, label) in enumerate(zip(x, f, labels)):
    line = Line2D((xi, xi), (0, fi), color="grey", linestyle="--")
    ax.add_line(line)

    txt = ax.text(xi, fi+0.01, label.replace("x", "f"), horizontalalignment="right", verticalalignment="bottom")

    if i >= 3:
        txt.set_horizontalalignment("center")

    if i >= 5:
        txt.set_horizontalalignment("left")


dx = x[1] - x[0]
arrow = FancyArrowPatch((x[1], f[1] - 0.1), (x[2], f[1] - 0.1), arrowstyle='<->', mutation_scale=20)
ax.add_patch(arrow)
ax.text(x[1] + dx/2, f[1] - 0.1, r"$\Delta x$", horizontalalignment="center", verticalalignment="bottom")

ax.spines["right"].set_visible(False)
ax.spines["top"].set_visible(False)

bottom = ax.yaxis.get_view_interval()[0]
ax.plot(x, [bottom]*len(x), 'o', clip_on=False, color="C0")

plt.savefig("func_on_grid.png")

plt.show()




import numpy as np

# import matplotlib.pyplot as plt
from numpy import cos, sin
from numpy.testing import assert_array_almost_equal

from findiff import Diff, Identity


def test_partial_d_dx():
    shape = (101,)
    x, dx = make_grid(shape, (0, 1))

    u = x**2
    expected = 2 * x

    fd = Diff(0, dx)
    actual = fd(u)

    assert_array_almost_equal(expected, actual)


def test_partial_d2_dx2():
    shape = (101,)
    x, dx = make_grid(shape, (0, 1))

    u = x**2
    expected = 2

    fd = Diff(0, dx) ** 2

    actual = fd(u, dx)

    assert_array_almost_equal(expected, actual)


def test_partial_d_dx_acc():
    shape = (11,)
    x, dx = make_grid(shape, (0, 1))

    u = x**3
    expected = 3 * x**2

    fd = Diff(0, dx)
    actual = fd(u)
    np.testing.assert_raises(
        AssertionError, assert_array_almost_equal, expected, actual
    )

    actual = fd(u, acc=4)
    assert_array_almost_equal(expected, actual)


def test_partial_d2_dxdy():

    shape = 50, 50
    (x, y), (dx, dy), (X, Y) = make_grid(shape, ((0, 1), (0, 1)))
    u = np.sin(X) * np.sin(Y)

    fd = Diff(0) * Diff(1)
    fd.set_grid({0: dx, 1: dy})
    actual = fd(u)

    expected = np.cos(X) * np.cos(Y)
    assert_array_almost_equal(expected, actual, decimal=3)


def test_partial_d_dx_on_2d_array():

    shape = 50, 50
    (x, y), (dx, dy), (X, Y) = make_grid(shape, ((0, 1), (0, 1)))
    u = np.sin(X) * np.sin(Y)

    fd = Diff(1, dy)
    actual = fd(u)

    expected = np.sin(X) * np.cos(Y)
    assert_array_almost_equal(expected, actual, decimal=3)


def test_add_two_diffs():

    shape = 50, 50
    (x, y), (dx, dy), (X, Y) = make_grid(shape, ((0, 1), (0, 1)))
    u = np.sin(X) * np.sin(Y)

    fd = Diff(0, dx) ** 2 + Diff(1, dy) ** 2
    actual = fd(u)

    expected = -2 * np.sin(X) * np.sin(Y)
    assert_array_almost_equal(expected, actual, decimal=3)


def test_sub_two_diffs():

    shape = 50, 50
    (x, y), (dx, dy), (X, Y) = make_grid(shape, ((0, 1), (0, 1)))
    u = np.sin(X) * np.sin(Y)

    fd = Diff(0, dx) ** 2 - Diff(1, dy) ** 2
    actual = fd(u)

    expected = np.zeros(shape)
    assert_array_almost_equal(expected, actual, decimal=3)


def test_multiply_with_scalar():

    shape = 50, 50
    (x, y), (dx, dy), (X, Y) = make_grid(shape, ((0, 1), (0, 1)))
    u = np.sin(X) * np.sin(Y)

    fd = 3 * Diff(0, dx) ** 2
    actual = fd(u, acc=4)

    expected = -3 * np.sin(X) * np.sin(Y)
    assert_array_almost_equal(expected, actual)


def test_multiply_with_variable_from_left():

    shape = 50, 50
    (x, y), (dx, dy), (X, Y) = make_grid(shape, ((0, 1), (0, 1)))
    u = np.sin(X) * np.sin(Y)

    fd = 3 * X * Y * Diff(0, dx) ** 2
    actual = fd(u, acc=4)

    expected = -3 * X * Y * np.sin(X) * np.sin(Y)
    assert_array_almost_equal(expected, actual)


def test_identity():

    shape = 50, 50
    (x, y), (dx, dy), (X, Y) = make_grid(shape, ((0, 1), (0, 1)))
    u = np.sin(X) * np.sin(Y)

    fd = Diff(0, dx) ** 2 + Identity()
    actual = fd(u, acc=4)

    expected = np.zeros_like(u)
    assert_array_almost_equal(expected, actual)


def test_linear_comb():
    shape = 50, 50
    (x, y), (dx, dy), (X, Y) = make_grid(shape, ((0, 1), (0, 1)))
    u = np.sin(X) * np.sin(Y)

    fd = 3 * X * Diff(0, dx) ** 2 - Y * (Diff(1, dy) ** 2 + Diff(0, dx))
    actual = fd(u, acc=4)

    expected = -3 * X * u - Y * (-sin(X) * sin(Y) + cos(X) * sin(Y))
    assert_array_almost_equal(expected, actual)


def make_grid(shape, edges):

    if len(shape) == 1:
        x = np.linspace(*edges, shape[0])
        dx = x[1] - x[0]
        return x, dx

    axes = tuple(
        [np.linspace(edges[k][0], edges[k][1], shape[k]) for k in range(len(shape))]
    )
    coords = np.meshgrid(*axes, indexing="ij")
    spacings = [axes[k][1] - axes[k][0] for k in range(len(shape))]
    return axes, spacings, coords


def test_nonuniform_1d_different_accs():
    x = np.r_[np.arange(0, 4, 0.05), np.arange(4, 10, 1)]
    f = np.exp(-(x**2))

    d_dx = Diff(0, x, acc=4)
    f_x = d_dx(f)
    assert_array_almost_equal(-2 * x * np.exp(-(x**2)), f_x, decimal=4)

    # same, but this time with default acc
    x = np.linspace(0, 1, 100)
    f = np.exp(-(x**2))
    d_dx = Diff(0, x)
    f_x = d_dx(f)
    assert_array_almost_equal(-2 * x * np.exp(-(x**2)), f_x, decimal=4)


def test_3d_different_accs():
    x = np.r_[np.arange(0, 4, 0.05), np.arange(4, 10, 1)]
    y = np.r_[np.arange(0, 4, 0.05), np.arange(4, 10, 1)]
    z = np.r_[np.arange(0, 4.5, 0.05), np.arange(4.5, 10, 1)]
    X, Y, Z = np.meshgrid(x, y, z, indexing="ij")
    f = np.exp(-(X**2) - Y**2 - Z**2)

    d_dy = Diff(1, y, acc=4)
    fy = d_dy(f)
    fye = -2 * Y * np.exp(-(X**2) - Y**2 - Z**2)
    assert_array_almost_equal(fy, fye, decimal=4)

    d_dy = Diff(1, y, acc=6)
    fy = d_dy(f)
    assert_array_almost_equal(fy, fye, decimal=4)



import matplotlib.pyplot as plt
from matplotlib.patches import Circle
import numpy as np


def label(ax, xy, names, font_size=12):
    text = ""
    for i, off in enumerate(xy):
        sign = "-" if off < 0 else "+"
        if off == 0:
            sign = ""
        value = abs(off) if abs(off) > 0 else ""
        text += "%s%s%s" % (names[i], sign, value)
        if i != len(xy) - 1:
            text += ","

    y = xy[1] - 0.3
    ax.text(xy[0], y, text, ha="center", family='sans-serif', size=font_size)


def plot_stencil(ax, stencil, xrange=None, yrange=None, label=label, font_size=12):
    stencil = np.array(stencil)

    if xrange is None:
        xrange = get_range(stencil, 0)
    if yrange is None:
        yrange = get_range(stencil, 1)

    for s in stencil:
        ax.add_patch(Circle(s, 0.1))
        label(ax, s, "ij", font_size)

    ax.set_aspect('equal')
    ax.set_xlim(*xrange)
    ax.set_ylim(*yrange)
    ax.set_axis_off()
    ax.grid(visible=True, axis='both')


def plot_num_grid(ax, xrange, yrange):
    for x in range(*xrange):
        for y in range(*yrange):
            ax.add_patch(Circle((x, y), 0.1, facecolor='#DDDDDD'))


def get_range(stc, axis):
    ax_min = min(stc[:, axis])
    ax_max = max(stc[:, axis])
    return (ax_min - 1, ax_max + 1)

#stc = [(0, 0), (-1, 0), (1, 0), (0, -1), (0, 1)]

from itertools import product
stc = list(product([-2, -1, 0, 1, 2], repeat=2))

fig = plt.figure()
ax = fig.add_subplot(111)

plot_num_grid(ax, (-3, 4), (-3, 4))
plot_stencil(ax, stc, font_size=8)
plt.show()



import matplotlib.pyplot as plt
import matplotlib as mpl


def plot_stencil(ax, grid, stencil, labels, stubs=[-0.5, 0.5], coefs=None, no_x=False):

    markersize = 16
    downshift = -0.01

    if coefs:
        markersize = 24
        downshift = -0.02


    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)

    ax.set_ylim(-0.1, 0.1)

    for direction in ['bottom', 'top', 'left', 'right']:
        ax.spines[direction].set_visible(False)

    ax.add_line(mpl.lines.Line2D([min(grid) + stubs[0], max(grid)+ stubs[1]], [0, 0], color='black'))
    ax.plot(grid, [0] * len(grid), 'o', color='#C0C0C0', markersize=markersize)
    ax.plot(stencil, [0] * len(stencil), 'o', color='C0', markersize=markersize)

    if coefs:
        for x, coef in zip(stencil, coefs):
            ax.text(x, 0, "$%s$" % coef, fontsize="x-small",
                    horizontalalignment="center", verticalalignment="center", color="white")

    for x, label in zip(grid, labels):
        if no_x:
            txt = "$ " + label + " $"
        else:
            txt = "$x_{%s}$" % label
        txt = ax.text(x, downshift, txt, clip_on=True, verticalalignment="top", horizontalalignment="center")
        if no_x:
            txt.set_fontsize("x-small")
            txt.set_y(1.5*downshift)

plt.rcParams.update({'font.size': 20, "mathtext.fontset": "cm"})

fig, ax = plt.subplots(figsize=(10, 2), layout='constrained')
plot_stencil(ax, list(range(-3, 4)), list(range(-1, 2)), labels=["k-3", "k-2", "k-1", "k", "k+1", "k+2", "k+3"])
fig.savefig("stencil_1d_center.png")

fig, ax = plt.subplots(figsize=(10, 2), layout='constrained')
plot_stencil(ax, list(range(8)), list(range(4)), labels=["0", "1", "2", "3", "4", "5", "6", "7"], stubs=[0, 0.5])
fig.savefig("stencil_1d_forward.png")

fig, ax = plt.subplots(figsize=(10, 2))
plot_stencil(ax, list(range(-3, 4)), list(range(-1, 2)), labels=["k-3", "k-2", "k-1", "k", "k+1", "k+2", "k+3"],
             coefs=[1, -2, 1]
             )
fig.savefig("stencil_1d_center_with_weights.png")

fig, ax = plt.subplots(figsize=(5, 2))
plot_stencil(ax, list(range(-1, 2)), list(range(-1, 2)), labels=["-1", "0", "1"],
             coefs=[1, -2, 1], no_x=True
             )
fig.savefig("stencil_1d_center_compact.png")

plt.show()


import numbers
from abc import ABC, abstractmethod

import numpy as np
from scipy import sparse

from findiff.grids import EquidistantGrid, Grid, TensorProductGrid
from findiff.stencils import StencilSet


class Node(ABC):
    """Represents a differential operator expression."""

    __array_priority__ = 100  # Makes sure custom multiplication is called over numpy's

    def __init__(self, *args, **kwargs):
        self.children = []

    @abstractmethod
    def __call__(self, f, *args, **kwargs):
        """Applies the differential operator expression L to array f.

        Parameters
        ----------
        f : numpy.ndarray
            The arrow to differentiate.

        Returns
        -------
        numpy.ndarray
            Array of the same shape as f containing the derivative values.
        """
        pass

    @abstractmethod
    def matrix(self, shape):
        """Returns a matrix representation of the differential operator for a given grid shape."""
        pass

    def stencil(self, shape):
        """Returns a stencil representation of the differential operator for a given grid shape."""
        return StencilSet(self, shape)

    def __add__(self, other):
        """Allows to add differential operator expressions."""
        return Add(self, other)

    def __radd__(self, other):
        """Allows to add differential operator expressions."""
        return Add(self, other)

    def __sub__(self, other):
        """Allows to subtract differential operator expressions."""
        return Add(ScalarOperator(-1) * other, self)

    def __rsub__(self, other):
        """Allows to subtract differential operator expressions."""
        return Add(ScalarOperator(-1) * other, self)

    def __mul__(self, other):
        """Allows to multiply differential operator expressions."""
        return Mul(self, other)

    def __rmul__(self, other):
        """Allows to multiply differential operator expressions."""
        return Mul(other, self)

    @property
    def grid(self):
        return getattr(self, "_grid", None)

    def set_grid(self, grid):
        """Sets the grid for the given differential operator expression.

        Parameters
        ----------
        grid: dict | Grid
            Specifies the grid to use. If a dict is given, an equidistant grid
            is assumed and the dict specifies the spacings along the required axes.
        """
        if isinstance(grid, dict):
            self._grid = EquidistantGrid(grid)
            for child in self.children:
                child.set_grid(self._grid)
        else:
            # Grid may be set lazily
            self._grid = grid
            for child in self.children:
                child.set_grid(self._grid)

    def set_accuracy(self, acc):
        """Sets the requested accuracy for the given differential operator expression.

        Parameters
        ----------
        acc: int
            The accuracy order. Must be a positive, even number.
        """
        self.acc = acc
        for child in self.children:
            child.set_accuracy(acc)


class FieldOperator(Node):
    """An operator that multiplies an array pointwise."""

    def __init__(self, value):
        super().__init__()
        self.value = value

    def __call__(self, f, *args, **kwargs):
        if isinstance(f, (numbers.Number, np.ndarray)):
            return self.value * f
        return self.value * super().__call__(f, *args, **kwargs)

    def matrix(self, shape):
        if isinstance(self.value, np.ndarray):
            diag_values = self.value.reshape(-1)
            return sparse.diags(diag_values)
        elif isinstance(self.value, numbers.Number):
            siz = np.prod(shape)
            return sparse.diags(self.value * np.ones(siz))


class ScalarOperator(FieldOperator):
    """A multiple of the identity operator."""

    def __init__(self, value):
        if not isinstance(value, numbers.Number):
            raise ValueError("Expected number, got " + str(type(value)))
        super().__init__(value)

    def matrix(self, shape):
        siz = np.prod(shape)
        mat = sparse.lil_matrix((siz, siz))
        diag = list(range(siz))
        mat[diag, diag] = self.value
        return sparse.csr_matrix(mat)


class Identity(ScalarOperator):
    """The identity operator."""

    def __init__(self):
        super().__init__(1)


class BinaryOperation(Node):

    @property
    def left(self):
        return self.children[0]

    @property
    def right(self):
        return self.children[1]


class Add(BinaryOperation):
    def __init__(self, left, right):
        if isinstance(left, (numbers.Number, np.ndarray)):
            left = FieldOperator(left)
        if isinstance(right, (numbers.Number, np.ndarray)):
            right = FieldOperator(right)
        super().__init__()
        self.children = [left, right]

    def __call__(self, f, *args, **kwargs):
        return self.left(f, *args, **kwargs) + self.right(f, *args, **kwargs)

    def matrix(self, shape):
        return self.left.matrix(shape) + self.right.matrix(shape)


class Mul(BinaryOperation):
    def __init__(self, left, right):
        if isinstance(left, (numbers.Number, np.ndarray)):
            left = FieldOperator(left)
        if isinstance(right, (numbers.Number, np.ndarray)):
            right = FieldOperator(right)
        super().__init__()
        self.children = [left, right]

    def __call__(self, f, *args, **kwargs):
        return self.left(self.right(f, *args, **kwargs), *args, **kwargs)

    def matrix(self, shape):
        return self.left.matrix(shape) * self.right.matrix(shape)


class Diff(Node):
    """Represents a partial derivative (along one axis).

    For higher derivatives, exponentiate. For mixed partial derivatives, multiply. See
    examples below.

    Examples
    --------
    Set up grid (equidistant here):
    >>> import numpy as np
    >>> x = np.linspace(0, 10, 100)

    The array to differentiate
    >>> f = np.sin(x) # as an example

    # Define the first derivative:
    >>> from findiff import Diff
    >>> d_dx = Diff(0)
    >>> d_dx = d_dx.set_grid({0: x[1] - x[0]})

    # now apply it:
    >>> df_dx = d_dx(f)

    # The second derivative is constructed by exponentiation:
    >>> d2_dx2 = d_dx ** 2
    >>> d2f_dx2 = d2_dx2(f)

    In multiple dimensions with meshed grids, the usage is accordingly:
    >>> x = y = z = np.linspace(0, 10, 100)
    >>> dx = dy = dz = x[1] - x[0]
    >>> X, Y, Z = np.meshgrid(x, y, z, indexing='ij')
    >>> f = np.sin(X) * np.sin(Y) * np.sin(Z)
    >>> d_dx = Diff(0)
    >>> d_dy = Diff(1)
    >>> d_dz = Diff(2)
    >>> d3_dxdydz = d_dx * d_dy * d_dz
    >>> d3_dxdydz.set_grid({0: dx, 1: dy, 2: dz})
    >>> d3f_dxdydz = d3_dxdydz(f)
    """

    DEFAULT_ACC = 2

    def __init__(self, axis=0, grid=None, acc=DEFAULT_ACC):
        """Initializes a Diff instance.

        Parameters
        ----------
        axis: int
            The 0-based index of the axis along which to take the derivative.
        grid: (optional) float | numpy.ndarray
            Specifies the grid to use. A float value assumes an equidistant grid
            and denoted the grid spacing along the given axis. A 1-D numpy array
            assumes an non-equidistant (tensor product) grid and denotes the coordinates
            along the given axis.
        acc: (optional) int
            The accuracy order to use. Must be a positive even number.
        """
        super().__init__()

        if isinstance(grid, numbers.Number):
            grid = {axis: grid}
        elif hasattr(grid, "shape") and hasattr(grid, "__len__"):
            grid = TensorProductGrid({axis: grid})

        self.set_grid(grid)

        self.axis = axis
        self._order = 1
        self.acc = acc

        self._fd = None

    @property
    def order(self):
        return self._order

    def __call__(self, f, *args, **kwargs):
        """Applies the differential operator."""

        if "acc" in kwargs:
            # allow to pass down new accuracy deep in expression tree
            new_acc = kwargs["acc"]
            if new_acc != self.acc:
                self._fd = None
                self.set_accuracy(new_acc)

        if self._fd is None:
            self._build_differentiator()

        if isinstance(f, Node):
            f = f(*args, **kwargs)
        return self._fd(f)

    def _build_differentiator(self):
        from findiff.legacy.operators import _FinDiff

        if isinstance(self.grid, EquidistantGrid):
            spacing = self.grid.spacing[self.axis]
            self._fd = _FinDiff(self.axis, spacing, self.order, acc=self.acc)
        elif isinstance(self.grid, TensorProductGrid):
            coords = self.grid.coords[self.axis]
            self._fd = _FinDiff(self.axis, coords, self.order, acc=self.acc)

    def matrix(self, shape):
        if not self._fd:
            self._build_differentiator()
        return self._fd.matrix(shape)

    def __pow__(self, power):
        """Returns a Diff instance for a higher order derivative."""
        new_diff = Diff(self.axis, acc=self.acc, grid=self.grid)
        new_diff._order *= power
        return new_diff



import numpy as np

from findiff import Diff, BoundaryConditions, PDE, Identity
from tests.test_diff import make_grid


def test_1d_dirichlet_hom():

    shape = (11,)

    x = np.linspace(0, 1, 11)
    dx = x[1] - x[0]
    L = Diff(0, dx) ** 2

    bc = BoundaryConditions(shape)

    bc[0] = 1
    bc[-1] = 2

    pde = PDE(L, np.zeros_like(x), bc)
    u = pde.solve()
    expected = x + 1
    np.testing.assert_array_almost_equal(expected, u)


def test_1d_dirichlet_inhom():

    nx = 21
    shape = (nx,)

    x = np.linspace(0, 1, nx)
    dx = x[1] - x[0]
    L = Diff(0, dx) ** 2

    bc = BoundaryConditions(shape)

    bc[0] = 1
    bc[-1] = 2

    pde = PDE(L, 6 * x, bc)

    u = pde.solve()
    expected = x**3 + 1
    np.testing.assert_array_almost_equal(expected, u)


def test_1d_neumann_hom():
    nx = 11
    shape = (nx,)

    x = np.linspace(0, 1, nx)
    dx = x[1] - x[0]
    L = Diff(0, dx) ** 2

    bc = BoundaryConditions(shape)

    bc[0] = 1
    bc[-1] = Diff(0, dx), 2

    pde = PDE(L, np.zeros_like(x), bc)
    u = pde.solve()
    expected = 2 * x + 1
    np.testing.assert_array_almost_equal(expected, u)


def test_2d_dirichlet_hom():
    shape = (11, 11)

    x, y = np.linspace(0, 1, shape[0]), np.linspace(0, 1, shape[1])
    dx, dy = x[1] - x[0], y[1] - y[0]
    X, Y = np.meshgrid(x, y, indexing="ij")
    L = Diff(0, dx) ** 2 + Diff(1, dy) ** 2

    expected = X + 1

    bc = BoundaryConditions(shape)

    bc[0, :] = 1
    bc[-1, :] = 2
    bc[:, 0] = X + 1
    bc[:, -1] = X + 1

    pde = PDE(L, np.zeros_like(X), bc)
    u = pde.solve()

    np.testing.assert_array_almost_equal(expected, u)


def test_2d_dirichlet_inhom():
    shape = (11, 11)

    x, y = np.linspace(0, 1, shape[0]), np.linspace(0, 1, shape[1])
    dx, dy = x[1] - x[0], y[1] - y[0]
    X, Y = np.meshgrid(x, y, indexing="ij")
    L = Diff(0, dx) ** 2 + Diff(1, dy) ** 2

    expected = X**3 + Y**3 + 1
    f = 6 * X + 6 * Y

    bc = BoundaryConditions(shape)

    bc[0, :] = expected
    bc[-1, :] = expected
    bc[:, 0] = expected
    bc[:, -1] = expected

    pde = PDE(L, f, bc)
    u = pde.solve()
    np.testing.assert_array_almost_equal(expected, u)


def test_2d_neumann_hom():
    shape = (31, 31)

    x, y = np.linspace(0, 1, shape[0]), np.linspace(0, 1, shape[1])
    dx, dy = x[1] - x[0], y[1] - y[0]
    X, Y = np.meshgrid(x, y, indexing="ij")
    L = Diff(0, dx) ** 2 + Diff(1, dy) ** 2

    expected = X**2 + Y**2 + 1
    f = 4 * np.ones_like(X)

    bc = BoundaryConditions(shape)

    d_dy = Diff(1, dy)

    bc[0, :] = expected
    bc[-1, :] = expected
    bc[:, 0] = d_dy, 2 * Y
    bc[:, -1] = d_dy, 2 * Y

    pde = PDE(L, f, bc)
    u = pde.solve()
    np.testing.assert_array_almost_equal(expected, u)


def test_1d_oscillator_free_dirichlet():
    n = 300
    shape = (n,)
    t = np.linspace(0, 5, n)
    dt = t[1] - t[0]
    L = Diff(0, dt) ** 2 + Identity()

    bc = BoundaryConditions(shape)

    bc[0] = 1
    bc[-1] = 2

    eq = PDE(L, np.zeros_like(t), bc)
    u = eq.solve()
    expected = np.cos(t) - (np.cos(5) - 2) * np.sin(t) / np.sin(5)
    np.testing.assert_array_almost_equal(expected, u, decimal=4)


def test_1d_damped_osc_driv_dirichlet():
    n = 100
    shape = (n,)
    t = np.linspace(0, 1, n)
    dt = t[1] - t[0]
    L = Diff(0, dt) ** 2 - Diff(0, dt) + Identity()
    f = -3 * np.exp(-t) * np.cos(t) + 2 * np.exp(-t) * np.sin(t)

    expected = np.exp(-t) * np.sin(t)

    bc = BoundaryConditions(shape)

    bc[0] = expected[0]
    bc[-1] = expected[-1]

    eq = PDE(L, f, bc)
    u = eq.solve()

    np.testing.assert_array_almost_equal(expected, u, decimal=4)


def test_1d_oscillator_driv_neumann():
    n = 200
    shape = (n,)
    t = np.linspace(0, 1, n)
    dt = t[1] - t[0]
    L = Diff(0, dt) ** 2 - Diff(0, dt) + Identity()
    f = -3 * np.exp(-t) * np.cos(t) + 2 * np.exp(-t) * np.sin(t)

    expected = np.exp(-t) * np.sin(t)

    bc = BoundaryConditions(shape)

    bc[0] = Diff(0, dt), 1
    bc[-1] = expected[-1]

    eq = PDE(L, f, bc)
    u = eq.solve()

    np.testing.assert_array_almost_equal(expected, u, decimal=4)


def test_1d_with_coeffs():
    n = 200
    shape = (n,)
    t = np.linspace(0, 1, n)
    dt = t[1] - t[0]
    L = t * Diff(0, dt) ** 2
    f = 6 * t**2

    bc = BoundaryConditions(shape)

    bc[0] = 0
    bc[-1] = 1

    eq = PDE(L, f, bc)
    u = eq.solve()
    expected = t**3

    np.testing.assert_array_almost_equal(expected, u, decimal=4)


def test_mixed_equation__with_coeffs_2d():
    shape = (41, 51)

    x, y = np.linspace(0, 1, shape[0]), np.linspace(0, 1, shape[1])
    dx, dy = x[1] - x[0], y[1] - y[0]
    X, Y = np.meshgrid(x, y, indexing="ij")
    L = Diff(0, dx) ** 2 + X * Y * Diff(0, dx) * Diff(1, dy) + Diff(1, dy) ** 2

    expected = X**3 + Y**3 + 1
    f = 6 * (X + Y)

    bc = BoundaryConditions(shape)

    bc[0, :] = expected
    bc[-1, :] = expected
    bc[:, 0] = expected
    bc[:, -1] = expected

    pde = PDE(L, f, bc)
    u = pde.solve()
    np.testing.assert_array_almost_equal(expected, u, decimal=4)


def test_2d_inhom_const_coefs_dirichlet_all():
    shape = (41, 50)
    (x, y), (dx, dy), (X, Y) = make_grid(shape, edges=[(-1, 1), (-1, 1)])

    expected = X**3 + Y**3 + X * Y + 1

    L = 3 * Diff(0, dx) ** 2 + 2 * Diff(0, dx) * Diff(1, dy) + Diff(1, dy) ** 2
    f = 2 + 18 * X + 6 * Y

    bc = BoundaryConditions(shape)
    bc[0, :] = expected
    bc[-1, :] = expected
    bc[:, 0] = expected
    bc[:, -1] = expected

    pde = PDE(L, f, bc)
    actual = pde.solve()
    np.testing.assert_array_almost_equal(expected, actual, decimal=4)


def test_2d_inhom_var_coefs_dirichlet_all():
    shape = (41, 50)
    (x, y), (dx, dy), (X, Y) = make_grid(shape, edges=[(-1, 1), (-1, 1)])

    expected = X**3 + Y**3 + X * Y + 1

    L = 3 * X * Diff(0, dx) ** 2 + 2 * Y * Diff(0, dx) * Diff(1, dy) + Diff(1, dy) ** 2
    f = 18 * X**2 + 8 * Y

    bc = BoundaryConditions(shape)
    bc[0, :] = expected
    bc[-1, :] = expected
    bc[:, 0] = expected
    bc[:, -1] = expected

    pde = PDE(L, f, bc)
    actual = pde.solve()
    np.testing.assert_array_almost_equal(expected, actual, decimal=4)



import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.patheffects as path_effects
from itertools import product

from plt_grids_with_stencils import plot_axes

fig = plt.figure(figsize=(5, 5), layout="constrained")
ax = fig.add_axes([0.0, 0, 1, 1])
ax.margins(0.2, 0.2)


offsets = list(product([-1, 0, 1], repeat=2)) + [(-2, -2), (-2, 2), (2, 2), (2, -2)]
#offsets = list(product([-1, 0, 1], repeat=2)) + [(-2, 0), (2, 0), (0, 2), (0, -2)]

plot_axes(ax, offsets, None, [-2, -1, 0, 1, 2], with_coefs=False, with_offsets=False)
t = ax.text(
    0.5, 0.95, "findiff",
    path_effects=[path_effects.withSimplePatchShadow()],
    transform=fig.transFigure,
    fontsize=65,
    fontstyle="italic",
    fontweight="medium",
    fontfamily="serif",
    color="C0",
    horizontalalignment="center",
    verticalalignment="top"
)

t.set_bbox(dict(facecolor='white', alpha=1, edgecolor='white'))
t.set_path_effects([
    path_effects.PathPatchEffect(
        offset=(4, -4), hatch='.......', facecolor='#A0A0A0'),
    path_effects.PathPatchEffect(
        edgecolor='white', linewidth=1.1, facecolor='C0')])

fig.savefig("findiff_logo.png")

plt.show()


"""
findiff is a Python package for finite difference numerical derivatives
and partial differential equations in any number of dimensions.

Features:

- Differentiate arrays of any number of dimensions along any axis with any desired accuracy order
- Accurate treatment of grid boundary
- Includes standard operators from vector calculus like gradient, divergence and curl
- Can handle uniform and non-uniform grids
- Can handle arbitrary linear combinations of derivatives with constant and variable coefficients
- Fully vectorized for speed
- Calculate raw finite difference coefficients for any order and accuracy for uniform and non-uniform grids
- New in version 0.7: Generate matrix representations of arbitrary linear differential operators
- New in version 0.8: Solve partial differential equations with Dirichlet or Neumann boundary conditions
- New in version 0.9: Generate differential operators for generic stencils
- New in version 0.10: Create symbolic representations of finite difference schemes
- Version 1.0: Completely remodeled API (backward compatibility is maintained, though)
"""

__version__ = "0.11.1"


from .legacy import *
from .operators import Diff, Identity
from .pde import PDE, BoundaryConditions
from .compatible import Coef, Coefficient, FinDiff, Id
from .coefs import coefficients
from .vector import Gradient, Divergence, Curl, Laplacian



import math
from itertools import product

import numpy as np

from .utils import to_long_index, to_index_tuple


class StencilSet(object):
    """
    Represent the finite difference stencil for a given differential operator.
    """

    def __init__(self, diff_op, shape, old_stl=None):
        """
        Constructor for Stencil objects.

        :param shape: tuple of ints
            Shape of the grid on which the stencil should be applied.

        :param axis: int >= 0
            The coordinate axis along which to take the partial derivative.

        :param order: int > 0
            The order of the derivative.

        :param h: float
            The spacing of the (equidistant) grid

        """

        self.shape = shape
        self.diff_op = diff_op
        self.char_pts = self._det_characteristic_points()

        self.data = {}

        self._create_stencil()

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return str(self.data)

    def apply(self, u, idx0):
        """ Applies the stencil to a point in an equidistant grid.

        :param u: ndarray
            An array with the function to differentiate.

        :param idx0: int or tuple of ints
            The index of the grid point where to differentiate the function.

        :return:
            The derivative at the given point.
        """

        if not hasattr(idx0, '__len__'):
            idx0 = (idx0,)

        typ = []
        for axis in range(len(self.shape)):
            if idx0[axis] == 0:
                typ.append('L')
            elif idx0[axis] == self.shape[axis] - 1:
                typ.append('H')
            else:
                typ.append('C')
        typ = tuple(typ)

        stl = self.data[typ]

        idx0 = np.array(idx0)
        du = 0.
        for o, c in stl.items():
            idx = idx0 + o
            du += c * u[tuple(idx)]

        return du

    def apply_all(self, u):
        """ Applies the stencil to all grid points.

        :param u: ndarray
            An array with the function to differentiate.

        :return:
            An array with the derivative.
        """

        assert self.shape == u.shape

        ndims = len(u.shape)
        if ndims == 1:
            indices = list(range(len(u)))
        else:
            axes_indices = []
            for axis in range(ndims):
                axes_indices.append(list(range(u.shape[axis])))

            axes_indices = tuple(axes_indices)
            indices = list(product(*axes_indices))

        du = np.zeros_like(u)

        for idx in indices:
            du[idx] = self.apply(u, idx)

        return du

    def _create_stencil(self):

        matrix = self.diff_op.matrix(self.shape)

        for pt in self.char_pts:

            char_point_stencil = {}
            self.data[pt] = char_point_stencil

            index_tuple_for_char_pt = self._typical_index_tuple_for_char_point(pt)
            long_index_for_char_pt = to_long_index(index_tuple_for_char_pt, self.shape)

            row = matrix[long_index_for_char_pt, :]
            long_row_inds, long_col_inds = row.nonzero()

            for long_offset_ind in long_col_inds:
                offset_ind_tuple = np.array(to_index_tuple(long_offset_ind, self.shape), dtype=int)
                offset_ind_tuple -= np.array(index_tuple_for_char_pt, dtype=int)
                char_point_stencil[tuple(offset_ind_tuple)] = row[0, long_offset_ind]

    def _typical_index_tuple_for_char_point(self, pt):
        index_tuple_for_char_pt = []
        for axis, key in enumerate(pt):
            if key == 'L':
                index_tuple_for_char_pt.append(0)
            elif key == 'C':
                index_tuple_for_char_pt.append(self.shape[axis] // 2)
            else:
                index_tuple_for_char_pt.append(self.shape[axis] - 1)
        return tuple(index_tuple_for_char_pt)

    def _det_characteristic_points(self):
        shape = self.shape
        ndim = len(shape)
        typ = [("L", "C", "H")] * ndim
        return product(*typ)


class Stencil:

    def __init__(self, offsets, partials, spacings=None):

        self.partials = partials
        self.max_order = 100
        if not hasattr(offsets[0], "__len__"):
            ndims = 1
            self.offsets = [(off,) for off in offsets]
        else:
            ndims = len(offsets[0])
            self.offsets = offsets

        if spacings is None:
            spacings = [1] * ndims
        elif not hasattr(spacings, "__len__"):
            spacings = [spacings] * ndims
        assert len(spacings) == ndims
        self.spacings = spacings
        self.ndims = ndims
        self.sol, self.sol_as_dict = self._make_stencil()

    def __call__(self, f, at=None, on=None):
        if at is not None and on is None:
            return self._apply_at_single_point(f, at)
        if at is None and on is not None:
            if isinstance(on[0], slice):
                return self._apply_on_multi_slice(f, on)
            else:
                return self._apply_on_mask(f, on)
        raise Exception('Cannot specify both *at* and *on* parameters.')

    def __str__(self):
        return str(self.values)

    def __repr__(self):
        return str(self.values)

    def _apply_on_mask(self, f, mask):
        result = np.zeros_like(f)
        for offset, coeff in self.values.items():
            offset_mask = self._make_offset_mask(mask, offset)
            result[mask] += coeff * f[offset_mask]
        return result

    def _apply_on_multi_slice(self, f, on):
        result = np.zeros_like(f)
        base_mslice = [self._canonic_slice(sl, f.shape[axis]) for axis, sl in enumerate(on)]
        for off, coeff in self.values.items():
            off_mslice = list(base_mslice)
            for axis, off_ in enumerate(off):
                start = base_mslice[axis].start + off_
                stop = base_mslice[axis].stop + off_
                off_mslice[axis] = slice(start, stop)
            result[tuple(base_mslice)] += coeff * f[tuple(off_mslice)]
        return result

    def _apply_at_single_point(self, f, at):
        result = 0.
        at = np.array(at)
        for off, coeff in self.values.items():
            off = np.array(off)
            eval_at = at + off
            if np.any(eval_at < 0) or not np.all(eval_at < f.shape):
                raise Exception('Cannot evaluate outside of grid.')
            result += coeff * f[tuple(eval_at)]
        return result

    def _make_offset_mask(self, mask, offset):
        offset_mask = np.full_like(mask, fill_value=False, dtype=bool)
        mslice_off = []
        mslice_base = []
        for off_ in offset:
            if off_ == 0:
                sl_off = slice(None, None)
                sl_base = slice(None, None)
            elif off_ > 0:
                sl_off = slice(off_, None)
                sl_base = slice(None, -off_)
            else:
                sl_off = slice(None, off_)
                sl_base = slice(-off_, None)
            mslice_off.append(sl_off)
            mslice_base.append(sl_base)
        offset_mask[tuple(mslice_base)] = mask[tuple(mslice_off)]
        return offset_mask

    def _canonic_slice(self, sl, length):
        start = sl.start
        if start is None:
            start = 0
        if start < 0:
            start = length - start
        stop = sl.stop
        if stop is None:
            stop = 0
        if stop < 0:
            stop = length - start
        return slice(start, stop)

    @property
    def values(self):
        return self.sol_as_dict

    @property
    def accuracy(self):
        return self._calc_accuracy()

    def _calc_accuracy(self):
        tol = 1.E-6
        deriv_order = 0
        for pows in self.partials.keys():
            order = sum(pows)
            if order > deriv_order:
                deriv_order = order
        for order in range(deriv_order, deriv_order + 10):
            terms = self._multinomial_powers(order)
            for term in terms:
                row = self._system_matrix_row(term)
                resid = np.sum(np.array(self.sol) * np.array(row))
                if abs(resid) > tol and term not in self.partials:
                    return order - deriv_order

    def _make_stencil(self):
        sys_matrix, taylor_terms = self._system_matrix()
        rhs = [0] * len(self.offsets)

        for i, term in enumerate(taylor_terms):
            if term in self.partials:
                weight = self.partials[term]
                multiplicity = np.prod([math.factorial(a) for a in term])
                vol = np.prod([self.spacings[j] ** term[j] for j in range(self.ndims)])
                rhs[i] = weight * multiplicity / vol

        sol = np.linalg.solve(sys_matrix, rhs)
        assert len(sol) == len(self.offsets)
        return sol, {off: coef for off, coef in zip(self.offsets, sol) if coef != 0}

    def _system_matrix(self):
        rows = []
        used_taylor_terms = []
        for order in range(self.max_order):
            taylor_terms = self._multinomial_powers(order)
            for term in taylor_terms:
                rows.append(self._system_matrix_row(term))
                used_taylor_terms.append(term)
                if not self._rows_are_linearly_independent(rows):
                    rows.pop()
                    used_taylor_terms.pop()
                if len(rows) == len(self.offsets):
                    return np.array(rows), used_taylor_terms
        raise Exception('Not enough terms. Try to increase max_order.')

    def _system_matrix_row(self, powers):
        row = []
        for a in self.offsets:
            value = 1
            for i, power in enumerate(powers):
                value *= a[i] ** power
            row.append(value)
        return row

    def _multinomial_powers(self, the_sum):
        """Returns all tuples of a given dimension that add up to the_sum."""
        all_combs = list(product(range(the_sum + 1), repeat=self.ndims))
        return list(filter(lambda tpl: sum(tpl) == the_sum, all_combs))

    def _rows_are_linearly_independent(self, matrix):
        """Checks the linear independence of the rows of a matrix."""
        matrix = np.array(matrix).astype(float)
        return np.linalg.matrix_rank(matrix) == len(matrix)



"""Provides an interface to obsolete classes for backward compatibility."""

from findiff import Diff
from findiff.legacy.operators import _FinDiff
from findiff.operators import FieldOperator, Identity


def FinDiff(*args, **kwargs):

    if len(args) > 3:
        raise ValueError("FinDiff accepts not more than 3 positional arguments.")

    def diff_from_tuple(tpl):
        if len(tpl) == 3:
            axis, h, order = tpl
            return Diff(axis, h, **kwargs) ** order
        elif len(tpl) == 2:
            axis, h = tpl
            return Diff(axis, h, **kwargs)

    if isinstance(args[0], (list, tuple)):
        diffs = []
        for tpl in args:
            diffs.append(diff_from_tuple(tpl))
        fd = diffs[0]
        for diff in diffs[1:]:
            fd = fd * diff
        return fd

    return diff_from_tuple(args)


FinDiff.__doc__ = _FinDiff.__doc__


# Define aliasses for backward compatibility
Coefficient = Coef = FieldOperator
Id = Identity



import unittest

import numpy as np

from findiff import Identity, FinDiff

from findiff.stencils import Stencil


class TestStencilOperations(unittest.TestCase):

    def test_solve_laplace_2d_with_5_points(self):
        offsets = [(-1, 0), (0, 0), (1, 0), (0, 1), (0, -1)]

        stencil = Stencil(offsets, {(2, 0): 1, (0, 2): 1})

        expected = {(0, 0): -4, (-1, 0): 1, (1, 0): 1, (0, 1): 1, (0, -1): 1}

        self.assertEqual(expected, stencil.values)
        self.assertEqual(2, stencil.accuracy)

    def test_solve_laplace_2d_with_9_points(self):
        offsets = [
            (-1, 0),
            (0, 0),
            (1, 0),
            (0, 1),
            (0, -1),
            (-2, 0),
            (2, 0),
            (0, -2),
            (0, 2),
        ]

        stencil = Stencil(offsets, {(2, 0): 1, (0, 2): 1})

        expected = {
            (0, 0): -5,
            (-1, 0): 4 / 3.0,
            (1, 0): 4 / 3.0,
            (0, 1): 4 / 3.0,
            (0, -1): 4 / 3.0,
            (-2, 0): -1 / 12.0,
            (2, 0): -1 / 12.0,
            (0, -2): -1 / 12.0,
            (0, 2): -1 / 12.0,
        }

        self.assertEqual(4, stencil.accuracy)
        self.assertEqual(len(expected), len(stencil.values))
        for off, coeff in stencil.values.items():
            self.assertAlmostEqual(coeff, expected[off])

    def test_solve_laplace_2d_with_5_points_times_2(self):
        offsets = [(-1, 0), (0, 0), (1, 0), (0, 1), (0, -1)]

        stencil = Stencil(offsets, {(2, 0): 2, (0, 2): 2})

        expected = {(0, 0): -8, (-1, 0): 2, (1, 0): 2, (0, 1): 2, (0, -1): 2}

        self.assertEqual(expected, stencil.values)
        self.assertEqual(2, stencil.accuracy)

    def test_solve_laplace_2d_with_5_points_times_2_and_spacing(self):
        offsets = [(-1, 0), (0, 0), (1, 0), (0, 1), (0, -1)]

        stencil = Stencil(offsets, {(2, 0): 2, (0, 2): 2}, spacings=(0.1, 0.1))

        expected = {(0, 0): -800, (-1, 0): 200, (1, 0): 200, (0, 1): 200, (0, -1): 200}

        self.assertEqual(len(expected), len(stencil.values))
        for off, coeff in stencil.values.items():
            self.assertAlmostEqual(coeff, expected[off])
        self.assertEqual(2, stencil.accuracy)

    def test_apply_laplacian_laplacian(self):

        x = y = np.linspace(0, 1, 21)
        dx = dy = x[1] - x[0]
        X, Y = np.meshgrid(x, y, indexing="ij")

        f = X**3 + Y**3
        expected = 6 * X + 6 * Y

        offsets = [(-1, 0), (0, 0), (1, 0), (0, 1), (0, -1)]
        stencil = Stencil(offsets, {(2, 0): 1, (0, 2): 1}, spacings=(dx, dy))
        at = (3, 4)
        actual = stencil(f, at)
        self.assertAlmostEqual(expected[at], actual)

    def test_apply_laplacian_laplacian_stencil_x_form(self):

        x = y = np.linspace(0, 1, 21)
        dx = dy = x[1] - x[0]
        X, Y = np.meshgrid(x, y, indexing="ij")

        f = X**3 + Y**3
        expected = 6 * X + 6 * Y

        offsets = [(0, 0), (1, 1), (-1, -1), (1, -1), (-1, 1)]
        stencil = Stencil(offsets, {(2, 0): 1, (0, 2): 1}, spacings=(dx, dy))
        at = (3, 4)
        actual = stencil(f, at)
        self.assertAlmostEqual(expected[at], actual)

    def test_apply_laplacian_laplacian_stencil_outside_grid(self):

        x = y = np.linspace(0, 1, 21)
        dx = dy = x[1] - x[0]
        X, Y = np.meshgrid(x, y, indexing="ij")

        f = X**3 + Y**3

        offsets = [(-1, 0), (0, 0), (1, 0), (0, 1), (0, -1)]
        stencil = Stencil(offsets, {(2, 0): 1, (0, 2): 1}, spacings=(dx, dy))
        at = (0, 1)
        with self.assertRaises(Exception):
            stencil(f, at)

        at = (3, 20)
        with self.assertRaises(Exception):
            stencil(f, at)

    def test_apply_mixed_deriv(self):

        x = y = np.linspace(0, 1, 101)
        dx = dy = x[1] - x[0]
        X, Y = np.meshgrid(x, y, indexing="ij")

        f = np.exp(-(X**2) - Y**2)
        expected = 4 * X * Y * f

        offsets = [(0, 0), (1, 1), (-1, -1), (1, -1), (-1, 1)]
        stencil = Stencil(offsets, partials={(1, 1): 1}, spacings=(dx, dy))
        at = (3, 4)
        actual = stencil(f, at)
        self.assertAlmostEqual(expected[at], actual, places=5)

    def test_laplace_1d_9points(self):
        x = np.linspace(0, 1, 101)
        f = x**3
        expected = 6 * x
        offsets = list(range(-4, 5))
        stencil = Stencil(offsets, partials={(2,): 1}, spacings=(x[1] - x[0],))
        at = (8,)
        actual = stencil(f, at)
        self.assertAlmostEqual(expected[at], actual, places=5)

    def tests_apply_stencil_on_multislice(self):
        x = y = np.linspace(0, 1, 21)
        dx = dy = x[1] - x[0]
        X, Y = np.meshgrid(x, y, indexing="ij")

        f = X**3 + Y**3
        expected = 6 * X + 6 * Y

        offsets = [(-1, 0), (0, 0), (1, 0), (0, 1), (0, -1)]
        stencil = Stencil(offsets, {(2, 0): 1, (0, 2): 1}, spacings=(dx, dy))

        on_slice = slice(1, -1), slice(1, -1)
        actual = stencil(f, on=on_slice)
        np.testing.assert_array_almost_equal(expected[on_slice], actual[on_slice])

    def tests_apply_stencil_on_mask(self):
        x = y = np.linspace(0, 1, 21)
        dx = dy = x[1] - x[0]
        X, Y = np.meshgrid(x, y, indexing="ij")

        f = X**3 + Y**3
        expected = 6 * X + 6 * Y

        offsets = [(-1, 0), (0, 0), (1, 0), (0, 1), (0, -1)]
        stencil = Stencil(offsets, {(2, 0): 1, (0, 2): 1}, spacings=(dx, dy))

        mask = np.full_like(f, fill_value=False, dtype=bool)
        mask[1:-1, 1:-1] = True
        actual = stencil(f, on=mask)
        np.testing.assert_array_almost_equal(expected[mask], actual[mask])

    def test_helmholtz_stencil_issue_60(self):
        # This is a regression test for issue #60.

        H = Identity() - FinDiff(0, 1, 2)

        stencil_set = H.stencil((10,))

        expected = {
            ("L",): {(0,): -1.0, (1,): 5.0, (2,): -4.0, (3,): 1.0},
            ("C",): {(-1,): -1.0, (0,): 3.0, (1,): -1.0},
            ("H",): {(-3,): 1.0, (-2,): -4.0, (-1,): 5.0, (0,): -1.0},
        }

        actual = stencil_set.data
        self.assertEqual(len(expected), len(actual))
        self.assertEqual(expected.keys(), actual.keys())
        for key, expected_stencil in expected.items():
            actual_stencil = actual[key]

            self.assertEqual(expected_stencil.keys(), actual_stencil.keys())
            for offset, expected_coef in expected_stencil.items():
                actual_coef = actual_stencil[offset]
                self.assertAlmostEqual(expected_coef, actual_coef)



from abc import ABC


class Grid(ABC):
    pass


class EquidistantGrid(Grid):

    def __init__(self, spacing: dict):
        for axis, h in spacing.items():
            if h <= 0:
                raise ValueError("spacing must be greater than zero")

        # key -> value == axis -> spacing for axis
        self.spacing = spacing


class TensorProductGrid(Grid):

    def __init__(self, axes_coords: dict):
        if len(axes_coords) == 0:
            raise ValueError("axes_coords cannot be empty")
        for axis, c in axes_coords.items():
            if len(c.shape) != 1:
                raise ValueError(
                    "each item in axes_coords must have only one dimension"
                )
        self.coords = axes_coords



"""
This module determines finite difference coefficients for uniform and 
non-uniform grids for any desired even accuracy order.

Most important function:

coefficients(deriv, acc=None, offsets=None, symbolic=False)

to calculate the finite difference coefficients for a given derivative
order and given accuracy order to given offsets.
"""
from itertools import combinations

import math
import numpy as np
import sympy


def coefficients(deriv, acc=None, offsets=None, symbolic=False, analytic_inv=False):
    """
    Calculates the finite difference coefficients for given derivative order and accuracy order.

    If acc is given, the coefficients are calculated for central, forward and backward
    schemes resulting in the specified accuracy order.

    If offsets are given, the coefficients are calculated for the offsets as specified
    and the resulting accuracy order is computed.

    *Either* acc *or* offsets must be given.

    Assumes that the underlying grid is uniform. This function is available at the `findiff`
    package level.

    :param deriv: The derivative order.
    :type deriv: int > 0

    :param acc: The accuracy order.
    :type acc:  even int > 0:

    :param offsets: The offsets for which to calculate the coefficients.
    :type offsets: list of ints

    :raises ValueError: if invalid arguments are given

    :return: dict with the finite difference coefficients and corresponding offsets
    """

    _validate_deriv(deriv)

    if acc is not None and offsets:
        raise ValueError('acc and offsets cannot both be given')

    if offsets:
        if deriv >= len(offsets):
            raise ValueError(
                f'can not compute derivative of order {deriv} using {len(offsets)} offsets.'
            )
        return calc_coefs(deriv, offsets, symbolic, analytic_inv)

    if acc is None:
        raise ValueError('either acc or offsets has to be given')

    _validate_acc(acc)
    ret = {}

    # Determine central coefficients

    num_central = 2 * math.floor((deriv + 1) / 2) - 1 + acc
    num_side = num_central // 2
    offsets = list(range(-num_side, num_side+1))

    ret["center"] = calc_coefs(deriv, offsets, symbolic, analytic_inv)

    # Determine forward coefficients

    if deriv % 2 == 0:
        num_coef = num_central + 1
    else:
        num_coef = num_central

    offsets = list(range(num_coef))
    ret["forward"] = calc_coefs(deriv, offsets, symbolic, analytic_inv)

    # Determine backward coefficients

    offsets = list(range(-num_coef+1, 1))
    ret["backward"] = calc_coefs(deriv, offsets, symbolic, analytic_inv)

    return ret


def compute_inverse_Vandermonde(column, offsets, symbolic):
    """Computes a given column of the inverse of the Vandermonde matrix that
    belongs to given offsets, multiplied by the factorial of the column index.

    This code implements the first equation for b_kj under "Proof 1" in
    https://proofwiki.org/wiki/Inverse_of_Vandermonde_Matrix
    Note that the result has to be transposed because the original Vandermonde
    matrix is defined transposed as compared to _build_matrix.

    Also, we have 0-based indexing here, whereas the formulae in the wiki entry
    are 1-based.
    """
    if symbolic:
        take = lambda arr, ids: [arr[idx] for idx in ids]
        prod = sympy.prod
        minus = lambda x, arr: [x - val for val in arr]
    else:
        take = lambda arr, ids: arr[ids]
        prod = np.prod
        offsets = np.array(offsets)
        minus = lambda x, arr: x - arr

    n = len(offsets)
    k = column + 1
    inv_vandermonde_column = []
    if k==n:
        # If the number of offsets matches the derivative order + 1, there is a special
        # case, compare the lower part of the bracket in the equation in proofwiki.
        for j in range(n):
            denom = prod(minus(offsets[j], offsets[:j])) * prod(minus(offsets[j], offsets[j+1:]))
            inv_vandermonde_column.append(1 / denom)
    else:
        # This is the "regular" part of the bracket. First compute the sign that is the 
        # same for all entries in the column that we compute
        sign = (-1) ** (n-k)
        for j in range(n):
            # All indices except j
            range_wo_j = list(range(j)) + list(range(j+1, n))
            # Get all combinations of n-k indices that are ascending and do not contain j
            index_set = combinations(range_wo_j, r=n-k)
            enumerator = sum((prod(take(offsets, list(m))) for m in index_set))
            denominator = prod(minus(offsets[j], take(offsets, range_wo_j)))
            inv_vandermonde_column.append(sign * enumerator / denominator)

    if symbolic:
        fact = math.factorial(column)
        return [val * fact for val in inv_vandermonde_column]

    return np.array(inv_vandermonde_column) * math.factorial(column)

def calc_coefs(deriv, offsets, symbolic=False, analytic_inv=False):
    if analytic_inv:
        coefs = compute_inverse_Vandermonde(deriv, offsets, symbolic)
    else:
        matrix = _build_matrix(offsets, symbolic)
        rhs = _build_rhs(offsets, deriv, symbolic)
        if symbolic:
            coefs = sympy.linsolve((matrix, rhs))
            coefs = list(tuple(coefs)[0])
        else:
            coefs = np.linalg.solve(matrix, rhs)

    acc = _calc_accuracy(offsets, coefs, deriv, symbolic)

    if not symbolic:
        offsets = np.array(offsets)

    return {
        "coefficients": coefs,
        "offsets": offsets,
        "accuracy": acc
    }

def coefficients_non_uni(deriv, acc, coords, idx):
    """
    Calculates the finite difference coefficients for given derivative order and accuracy order.
    Assumes that the underlying grid is non-uniform.

    :param deriv: int > 0: The derivative order.

    :param acc:  even int > 0: The accuracy order.
     
    :param coords:  1D numpy.ndarray: the coordinates of the axis for the partial derivative
    
    :param idx:  int: index of the grid position where to calculate the coefficients

    :return: dict with the finite difference coefficients and corresponding offsets 
    """

    _validate_deriv(deriv)
    _validate_acc(acc)

    num_central = 2 * math.floor((deriv + 1) / 2) - 1 + acc
    num_side = num_central // 2

    if deriv % 2 == 0:
        num_coef = num_central + 1
    else:
        num_coef = num_central

    if idx < num_side:
        matrix = _build_matrix_non_uniform(0, num_coef - 1, coords, idx)

        offsets = list(range(num_coef))
        rhs = _build_rhs(offsets, deriv)

        ret = {
            "coefficients": np.linalg.solve(matrix, rhs),
            "offsets": np.array(offsets)
        }

    elif idx >= len(coords) - num_side:
        matrix = _build_matrix_non_uniform(num_coef - 1, 0, coords, idx)

        offsets = list(range(-num_coef+1, 1))
        rhs = _build_rhs(offsets, deriv)

        ret = {
            "coefficients": np.linalg.solve(matrix, rhs),
            "offsets": np.array(offsets)
        }

    else:
        matrix = _build_matrix_non_uniform(num_side, num_side, coords, idx)

        offsets = list(range(-num_side, num_side+1))
        rhs = _build_rhs(offsets, deriv)

        ret = {
            "coefficients": np.linalg.solve(matrix, rhs),
            "offsets": np.array([p for p in range(-num_side, num_side + 1)])
        }

    return ret


def _build_matrix(offsets, symbolic=False):
    """Constructs the equation system matrix for the finite difference coefficients"""

    A = [([1 for _ in offsets])]
    for i in range(1, len(offsets)):
        A.append([j**i for j in offsets])
    if symbolic:
        return sympy.Matrix(A)
    else:
        return np.array(A,dtype='float')


def _build_rhs(offsets, deriv, symbolic=False):
    """The right hand side of the equation system matrix"""

    b = [0 for _ in offsets]
    b[deriv] = math.factorial(deriv)
    if symbolic:
        return sympy.Matrix(b)
    else:
        return np.array(b,dtype='float')


def _build_matrix_non_uniform(p, q, coords, k):
    """Constructs the equation matrix for the finite difference coefficients of non-uniform grids at location k"""
    A = [[1] * (p+q+1)]
    for i in range(1, p + q + 1):
        line = [(coords[k+j] - coords[k])**i for j in range(-p, q+1)]
        A.append(line)
    return np.array(A,dtype='float')


def _calc_accuracy(offsets, coefs, deriv, symbolic=False):

    n = deriv + 1
    max_n = 999
    if symbolic:
        break_cond = lambda b: b != 0
    else:
        break_cond = lambda b: abs(b) > 1.E-6

    while True:
        b = 0
        #for i, coef in enumerate(coefs):
        for o, coef in zip(offsets, coefs):
            #k = min(offsets) + i
            b += coef * o ** n

        if break_cond(b):
            break

        n += 1
        if n > max_n:
            raise Exception('Cannot compute accuracy.')

    return round(n - deriv)


def _validate_acc(acc):
    if acc % 2 == 1 or acc <= 0:
        raise ValueError('Accuracy order acc must be positive EVEN integer')


def _validate_deriv(deriv):
    if deriv < 0:
        raise ValueError('Derive degree must be positive integer')



import numpy as np
from numpy.testing import assert_array_almost_equal

from findiff.operators import Add, FieldOperator
from findiff import Diff, Identity


def test_applying_identity():

    f = np.array([1, 2, 3])
    I = Identity()

    assert np.array_equal(I(f), f)


def test_add_identities():

    f = np.array([1, 2, 3])
    I = Identity()

    assert np.array_equal((I + I + I)(f), 3 * f)


def test_add_scalar_to_identity():

    f = np.array([1, 2, 3])
    I = Identity()

    Ip1 = I + 2

    assert isinstance(Ip1, Add)
    assert isinstance(Ip1.left, Identity)
    assert isinstance(Ip1.right, FieldOperator)
    assert np.array_equal(Ip1(f), 3 * f)

    Ip1 = 2 + I

    assert isinstance(Ip1, Add)
    assert isinstance(Ip1.left, Identity)
    assert isinstance(Ip1.right, FieldOperator)
    assert np.array_equal(Ip1(f), 3 * f)


def test_add_scalars():

    f = np.array([1, 2, 3])
    S = FieldOperator(2)

    assert np.array_equal((S + S)(f), 4 * f)


def test_sub_scalars():

    f = np.array([1, 2, 3])
    S = FieldOperator(2)

    actual = (S - 2 * S)(f)
    expected = -2 * f
    np.testing.assert_array_almost_equal(actual, expected)


def test_multiply_scalars():

    f = np.array([1, 2, 3])
    S = FieldOperator(3)

    assert np.array_equal((S * S)(f), 9 * f)


def test_multiply_scalar_with_number():

    f = np.array([1, 2, 3])
    S = FieldOperator(3)

    assert np.array_equal((S * 2)(f), 6 * f)
    assert np.array_equal((2 * S)(f), 6 * f)


def test_multiply_scalar_with_field():

    f = np.array([1, 2, 3])
    g = np.array([-1, 1, -1])
    S = FieldOperator(3)

    assert np.array_equal((S * g)(f), [-3, 6, -9])
    assert np.array_equal((g * S)(f), [-3, 6, -9])


def test_mix_addition_and_multiplication():
    f = np.array([1, 2, 3])
    S2 = FieldOperator(2)
    S3 = FieldOperator(3)

    assert np.array_equal((S2 + S3 * 4)(f), 2 * f + 3 * 4 * f)


def test_simple_derivatives():

    x = np.linspace(0, 1, 100)
    dx = x[1] - x[0]
    f = x**3

    d_dx = Diff(0, dx)

    actual = d_dx(f)
    np.testing.assert_array_almost_equal(actual, 3 * x**2, decimal=3)

    actual = d_dx(f, acc=4)
    np.testing.assert_array_almost_equal(actual, 3 * x**2)

    d_dx = Diff(0, dx)

    actual = d_dx(f, acc=4)
    np.testing.assert_array_almost_equal(actual, 3 * x**2)

    d2_dx2 = Diff(0, dx) ** 2

    actual = d2_dx2(f, acc=4)
    np.testing.assert_array_almost_equal(actual, 6 * x)


def test_pow_diff():

    d = Diff(0)
    assert d.order == 1

    d2 = d**2

    assert d2.order == 2
    assert d.order == 1


def test_chained_derivatives():

    x = np.linspace(0, 1, 100)
    dx = x[1] - x[0]
    f = x**3

    d_dx = Diff(0, dx)

    d2_dx2 = d_dx * d_dx
    actual = d2_dx2(f, acc=4)
    np.testing.assert_array_almost_equal(actual, 6 * x)


def test_set_grid_lazily():

    x = np.linspace(0, 1, 100)
    dx = x[1] - x[0]
    f = x**3

    # no grid defined here:
    d_dx = Diff(0)

    diff_op = d_dx * d_dx + d_dx

    # now set grid for complete differential operator
    diff_op.set_grid({0: dx})

    actual = diff_op(f, acc=4)
    np.testing.assert_array_almost_equal(actual, 6 * x + 3 * x**2)


def test_harmonic():

    x = np.linspace(0, 1, 100)
    dx = x[1] - x[0]
    f = x**3

    T = -0.5 * Diff(0, dx) ** 2
    V = 0.5 * x**2
    H = T + V

    actual = H(f, acc=4)
    Tf = T(f, acc=4)
    Vf = FieldOperator(V)(f)
    np.testing.assert_array_almost_equal(Tf, -3 * x)
    np.testing.assert_array_almost_equal(Vf, 0.5 * x**2 * f)
    np.testing.assert_array_almost_equal(actual, -3 * x + 0.5 * x**2 * f)


def test_partial_diff():
    nx = 100
    x = np.linspace(0, np.pi, nx)
    u = np.sin(x)
    ux_ex = np.cos(x)
    dx = x[1] - x[0]
    fd = Diff(0, dx, acc=4)

    ux = fd(u)

    assert_array_almost_equal(ux, ux_ex, decimal=5)

    ny = 80
    y = np.linspace(0, np.pi, ny)
    dy = y[1] - y[0]
    X, Y = np.meshgrid(x, y, indexing="ij")

    u = np.sin(X) * np.sin(Y)
    uxy_ex = np.cos(X) * np.cos(Y)

    # mixed partial derivative d2_dxdy
    fd = Diff(0, dx) * Diff(1, dy)

    uxy = fd(u, acc=4)

    assert_array_almost_equal(uxy, uxy_ex, decimal=5)


def test_matrix_1d():

    x = np.linspace(0, 6, 7)
    d2_dx2 = Diff(0, x[1] - x[0]) ** 2
    u = x**2

    mat = d2_dx2.matrix(u.shape)

    np.testing.assert_array_almost_equal(2 * np.ones_like(x), mat.dot(u.reshape(-1)))


def test_matrix_2d():
    thr = np.get_printoptions()["threshold"]
    lw = np.get_printoptions()["linewidth"]
    np.set_printoptions(threshold=np.inf)
    np.set_printoptions(linewidth=500)
    x, y = [np.linspace(0, 4, 5)] * 2
    X, Y = np.meshgrid(x, y, indexing="ij")
    laplace = Diff(0, x[1] - x[0]) ** 2 + Diff(0, y[1] - y[0]) ** 2
    # d = FinDiff(1, y[1]-y[0], 2)
    u = X**2 + Y**2

    mat = laplace.matrix(u.shape)

    np.testing.assert_array_almost_equal(
        4 * np.ones_like(X).reshape(-1), mat.dot(u.reshape(-1))
    )

    np.set_printoptions(threshold=thr)
    np.set_printoptions(linewidth=lw)


def test_matrix_2d_mixed():
    x, y = [np.linspace(0, 5, 6), np.linspace(0, 6, 7)]
    X, Y = np.meshgrid(x, y, indexing="ij")
    d2_dxdy = Diff(0, x[1] - x[0]) * Diff(1, y[1] - y[0])
    u = X**2 * Y**2

    mat = d2_dxdy.matrix(u.shape)
    expected = d2_dxdy(u).reshape(-1)

    actual = mat.dot(u.reshape(-1))
    np.testing.assert_array_almost_equal(expected, actual)


def test_matrix_1d_coeffs():
    shape = (11,)
    x = np.linspace(0, 10, 11)
    dx = x[1] - x[0]

    L = x * Diff(0, dx) ** 2

    u = np.random.rand(*shape).reshape(-1)

    actual = L.matrix(shape).dot(u)
    expected = L(u).reshape(-1)
    np.testing.assert_array_almost_equal(expected, actual)



import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
from itertools import product


def switch_off_xy_axis(ax):
    ax.xaxis.set_ticks([])
    ax.yaxis.set_ticks([])
    for direction in ['bottom', 'top', 'left', 'right']:
        ax.spines[direction].set_visible(False)


def plot_grid(ax, offsets, color, markersize=30):
    offsets = np.array(offsets)
    ax.plot(offsets[:, 0], offsets[:, 1], 'o', color=color, markersize=markersize)


def plot_offsets(ax, offsets, vshift=-0.15):
    for x, y in offsets:
        ax.text(x, y + vshift, "$ " + str((x, y)) + " $", verticalalignment="top", horizontalalignment="center")


def plot_coefficients(ax, offsets, coefs):
    if not coefs:
        return
    for (x, y), c in zip(offsets, coefs):
        ax.text(
            x, y, c,
            color="white",
            verticalalignment="center", horizontalalignment="center"
        )


def tex(s):
    return "$ " + str(s) + " $"


def plot_axes(ax, stencil, coeffs, grid_kernel=[-1, 0, 1], with_coefs=True, with_offsets=True, markersize=30):
    switch_off_xy_axis(ax)
    plot_grid(ax, list(product(grid_kernel, repeat=2)), color="#C0C0C0", markersize=markersize)
    plot_grid(ax, stencil, color="C0", markersize=markersize)
    if with_offsets:
        plot_offsets(ax, stencil)
    if with_coefs:
        plot_coefficients(ax, stencil, coeffs)


if __name__ == '__main__':

    plt.rcParams.update({'font.size': 20, "mathtext.fontset": "cm"})

    fig = plt.figure(figsize=(10, 10))

    ax = fig.add_axes([0, 0.5, 0.5, 0.5])
    ax.margins(0.4, 0.4)
    plot_axes(ax, stencil=[(-1, 0), (0, 0), (1, 0)], coeffs=[1, -2, 1])

    ax = fig.add_axes([0.5, 0.5, 0.5, 0.5])
    ax.margins(0.4, 0.4)
    plot_axes(ax, stencil=[(0, -1), (0, 0), (0, 1)], coeffs=[1, -2, 1])

    ax = fig.add_axes([0.25, 0, 0.5, 0.5])
    ax.margins(0.4, 0.4)
    plot_axes(ax, stencil=[(0, -1), (0, 0), (0, 1), (-1, 0), (1, 0)], coeffs=[1, -4, 1, 1, 1])

    fig.text(0.48, 0.72, "+", fontsize=34)
    fig.text(0.48, 0.5, "=", fontsize=34)

    fig.text(0.25, 0.95, r"$\frac{\partial^2}{\partial x^2}$", fontsize=30, horizontalalignment="center")
    fig.text(0.75, 0.95, r"$\frac{\partial^2}{\partial y^2}$", fontsize=30, horizontalalignment="center")

    fig.text(0.8, 0.25, r"$\frac{\partial^2}{\partial x^2} + \frac{\partial^2}{\partial y^2}$",
             fontsize=30, horizontalalignment="center", verticalalignment="center")

    fig.savefig("composite_stencil.png")

    plt.show()



import unittest

import numpy as np
from sympy import Rational

from findiff import coefficients
from findiff.coefs import calc_coefs
from findiff.coefs import coefficients_non_uni


class TestCoefs(unittest.TestCase):

    def test_order2_acc2(self):

        for analytic_inv in [True, False]:
            c = coefficients(deriv=2, acc=2, analytic_inv=analytic_inv)
            with self.subTest():

                coefs = c["center"]["coefficients"]
                np.testing.assert_array_almost_equal(np.array([1.0, -2.0, 1.0]), coefs)
                offs = c["center"]["offsets"]
                np.testing.assert_array_almost_equal(np.array([-1, 0, 1]), offs)

                coefs = c["forward"]["coefficients"]
                np.testing.assert_array_almost_equal(np.array([2, -5, 4, -1]), coefs)
                offs = c["forward"]["offsets"]
                np.testing.assert_array_almost_equal(np.array([0, 1, 2, 3]), offs)

                coefs = c["backward"]["coefficients"]
                np.testing.assert_array_almost_equal(
                    np.array(([2, -5, 4, -1])[::-1]), coefs
                )
                offs = c["backward"]["offsets"]
                np.testing.assert_array_almost_equal(np.array([-3, -2, -1, 0]), offs)

    def test_order1_acc2(self):

        for analytic_inv in [True, False]:
            c = coefficients(deriv=1, acc=2, analytic_inv=analytic_inv)
            with self.subTest():

                coefs = c["center"]["coefficients"]
                np.testing.assert_array_almost_equal(np.array([-0.5, 0, 0.5]), coefs)
                offs = c["center"]["offsets"]
                np.testing.assert_array_almost_equal(np.array([-1, 0, 1]), offs)

                coefs = c["forward"]["coefficients"]
                np.testing.assert_array_almost_equal(np.array([-1.5, 2, -0.5]), coefs)
                offs = c["forward"]["offsets"]
                np.testing.assert_array_almost_equal(np.array([0, 1, 2]), offs)

                coefs = c["backward"]["coefficients"]
                np.testing.assert_array_almost_equal(
                    -np.array([-1.5, 2, -0.5])[::-1], coefs
                )
                offs = c["backward"]["offsets"]
                np.testing.assert_array_almost_equal(np.array([-2, -1, 0]), offs)

    def test_order1_acc4(self):

        for analytic_inv in [True, False]:
            c = coefficients(deriv=1, acc=4, analytic_inv=analytic_inv)
            with self.subTest():
                coefs = c["center"]["coefficients"]
                np.testing.assert_array_almost_equal(
                    np.array([1 / 12, -2 / 3, 0, 2 / 3, -1 / 12]), coefs
                )
                offs = c["center"]["offsets"]
                np.testing.assert_array_almost_equal(np.array([-2, -1, 0, 1, 2]), offs)

                coefs = c["forward"]["coefficients"]
                np.testing.assert_array_almost_equal(
                    np.array([-25 / 12, 4, -3, 4 / 3, -1 / 4]), coefs
                )
                offs = c["forward"]["offsets"]
                np.testing.assert_array_almost_equal(np.array([0, 1, 2, 3, 4]), offs)

                coefs = c["backward"]["coefficients"]
                np.testing.assert_array_almost_equal(
                    -np.array([-25 / 12, 4, -3, 4 / 3, -1 / 4])[::-1], coefs
                )
                offs = c["backward"]["offsets"]
                np.testing.assert_array_almost_equal(
                    -np.array([0, 1, 2, 3, 4])[::-1], offs
                )

    def test_order2_acc4(self):

        for analytic_inv in [True, False]:
            c = coefficients(deriv=2, acc=4, analytic_inv=analytic_inv)
            with self.subTest():
                coefs = c["center"]["coefficients"]
                np.testing.assert_array_almost_equal(
                    np.array([-1 / 12, 4 / 3, -2.5, 4 / 3, -1 / 12]), coefs
                )
                offs = c["center"]["offsets"]
                np.testing.assert_array_almost_equal(np.array([-2, -1, 0, 1, 2]), offs)

                coefs = c["forward"]["coefficients"]
                np.testing.assert_array_almost_equal(
                    np.array([15 / 4, -77 / 6, 107 / 6, -13, 61 / 12, -5 / 6]), coefs
                )
                offs = c["forward"]["offsets"]
                np.testing.assert_array_almost_equal(np.array([0, 1, 2, 3, 4, 5]), offs)

                coefs = c["backward"]["coefficients"]
                np.testing.assert_array_almost_equal(
                    np.array([15 / 4, -77 / 6, 107 / 6, -13, 61 / 12, -5 / 6])[::-1],
                    coefs,
                )
                offs = c["backward"]["offsets"]
                np.testing.assert_array_almost_equal(
                    -np.array([0, 1, 2, 3, 4, 5])[::-1], offs
                )

    def test_calc_accuracy_central_deriv2_acc2(self):
        for analytic_inv in [True, False]:
            coefs = calc_coefs(2, [-1, 0, 1], analytic_inv=analytic_inv)
            with self.subTest():
                self.assertEqual(2, coefs["accuracy"])

    def test_calc_accuracy_central_deriv1_acc2(self):
        for analytic_inv in [True, False]:
            coefs = calc_coefs(1, [-1, 0, 1], analytic_inv=analytic_inv)
            with self.subTest():
                self.assertEqual(2, coefs["accuracy"])

    def test_calc_accuracy_left1_right0_deriv1_acc1(self):
        for analytic_inv in [True, False]:
            coefs = calc_coefs(1, [-1, 0], analytic_inv=analytic_inv)
            with self.subTest():
                self.assertEqual(1, coefs["accuracy"])

    def test_calc_accuracy_left0_right3_deriv1_acc3(self):
        for analytic_inv in [True, False]:
            coefs = calc_coefs(2, [0, 1, 2, 3], analytic_inv=analytic_inv)
            with self.subTest():
                self.assertEqual(2, coefs["accuracy"])

    def test_calc_accuracy_from_offsets_symbolic(self):
        for analytic_inv in [True, False]:
            coefs = calc_coefs(
                2, [0, 1, 2, 3], symbolic=True, analytic_inv=analytic_inv
            )
            with self.subTest():
                self.assertEqual(2, coefs["accuracy"])

    def test_calc_accuracy_from_offsets_symbolic(self):
        for analytic_inv in [True, False]:
            coefs = calc_coefs(
                2, [-4, -2, 0, 2, 4], symbolic=True, analytic_inv=analytic_inv
            )
            with self.subTest():
                self.assertEqual(4, coefs["accuracy"])

    def test_calc_coefs_from_offsets(self):
        for analytic_inv in [True, False]:
            coefs = calc_coefs(1, [-2, 0, 1], analytic_inv=analytic_inv)
            with self.subTest():
                np.testing.assert_array_almost_equal(
                    coefs["coefficients"], [-1.0 / 6, -0.5, 2.0 / 3]
                )

    def test_calc_coefs_from_offsets_no_central_point(self):
        for analytic_inv in [True, False]:
            coefs = calc_coefs(1, [-2, 1, 2], analytic_inv=analytic_inv)
            with self.subTest():
                np.testing.assert_array_almost_equal(
                    coefs["coefficients"], [-1.0 / 4, 0, 1.0 / 4]
                )

    def test_calc_coefs_symbolic(self):
        for analytic_inv in [True, False]:
            coefs = calc_coefs(1, [-2, 0, 1], symbolic=True, analytic_inv=analytic_inv)
            expected = [Rational(-1, 6), Rational(-1, 2), Rational(2, 3)]
            with self.subTest():
                np.testing.assert_array_almost_equal(coefs["coefficients"], expected)

    def test_non_uniform(self):

        x = np.linspace(0, 10, 100)
        dx = x[1] - x[0]

        for analytic_inv in [True, False]:
            c_uni = coefficients(deriv=2, acc=2, analytic_inv=analytic_inv)
            coefs_uni = c_uni["center"]["coefficients"] / dx**2

            c_non_uni = coefficients_non_uni(deriv=2, acc=2, coords=x, idx=5)
            coefs_non_uni = c_non_uni["coefficients"]

            with self.subTest():
                np.testing.assert_array_almost_equal(coefs_non_uni, coefs_uni)

    def test_invalid_acc_raises_exception(self):
        with self.assertRaises(ValueError):
            coefficients(deriv=1, acc=3)
        with self.assertRaises(ValueError):
            coefficients(deriv=1, acc=0)
        with self.assertRaises(ValueError):
            coefficients_non_uni(1, 3, None, None)
        with self.assertRaises(ValueError):
            coefficients_non_uni(1, 0, None, None)

    def test_invalid_deriv_raises_exception(self):
        with self.assertRaises(ValueError):
            coefficients(-1, 2)
        with self.assertRaises(ValueError):
            coefficients_non_uni(-1, 2, None, None)


if __name__ == "__main__":
    unittest.main()



import unittest

from sympy import IndexedBase, Symbol, symbols, latex

from findiff.symbolic import SymbolicMesh, SymbolicDiff


class TestSymbolicMesh(unittest.TestCase):

    def test_parse_symbolic_mesh(self):
        # 1D
        mesh = SymbolicMesh(coord="x", equidistant=True)
        (x,) = mesh.coord
        (dx,) = mesh.spacing

        self.assertEqual(IndexedBase, type(x))
        self.assertEqual(Symbol, type(dx))

        # 2D
        mesh = SymbolicMesh(coord="x,y", equidistant=True)
        x, y = mesh.coord
        dx, dy = mesh.spacing

        self.assertEqual(IndexedBase, type(x))
        self.assertEqual(IndexedBase, type(y))
        self.assertEqual(Symbol, type(dx))
        self.assertEqual(Symbol, type(dy))

        # ignores whitespace
        mesh = SymbolicMesh(coord="x,  y", equidistant=True)
        x, y = mesh.coord

        self.assertEqual("x", str(x))
        self.assertEqual("y", str(y))

    def test_create_symbol(self):
        # defaults
        mesh = SymbolicMesh(coord="x", equidistant=True)
        actual = mesh.create_symbol("u")
        expected = IndexedBase("u")
        self.assertEqual(latex(actual), latex(expected))

        # both indices down
        mesh = SymbolicMesh(coord="x,y", equidistant=True)
        n, m = symbols("n, m")
        actual = latex(mesh.create_symbol("u")[n, m])
        # expected = "u{}_{n}{}_{m}"
        expected = "{u}_{n,m}"
        self.assertEqual(latex(actual), latex(expected))

        # both indices up
        # mesh = SymbolicMesh(coord="x,y", equidistant=True)
        # n, m = symbols("n, m")
        # u = mesh.create_symbol("u", pos="u,u")
        # actual = latex(u[n, m])
        # expected = "u{}^{n}{}^{m}"
        # self.assertEqual(actual, expected)


class TestDiff(unittest.TestCase):

    def test_init(self):
        mesh = SymbolicMesh("x")
        d = SymbolicDiff(mesh)

        self.assertEqual(d.axis, 0)
        self.assertEqual(d.degree, 1)
        self.assertEqual(id(mesh), id(d.mesh))

    def test_call(self):
        # 1D
        mesh = SymbolicMesh("x")
        u = mesh.create_symbol("u")
        d = SymbolicDiff(mesh)
        n = Symbol("n")

        actual = d(u, at=n, offsets=[-1, 0, 1])

        expected = (u[n + 1] - u[n - 1]) / (2 * mesh.spacing[0])

        self.assertEqual(0, (expected - actual).simplify())

        # 2D
        mesh = SymbolicMesh("x, y")
        u = mesh.create_symbol("u")
        d = SymbolicDiff(mesh, axis=1)
        n, m = symbols("n, m")

        actual = d(u, at=(n, m), offsets=[-1, 0, 1])

        expected = (u[n, m + 1] - u[n, m - 1]) / (2 * mesh.spacing[1])

        self.assertEqual(0, (expected - actual).simplify())


if __name__ == "__main__":
    unittest.main()



import matplotlib.pyplot as plt

from plt_grids_with_stencils import *

plt.rcParams.update({'font.size': 20, "mathtext.fontset": "cm"})

fig, ax = plt.subplots(figsize=(8, 8))
plot_axes(ax, stencil=[(-1, 0), (0, 0), (1, 0), (0, 1), (0, -1)], coeffs=[1, -4, 1, 1, 1], markersize=40)
fig.savefig("laplace2d.png")

fig, ax = plt.subplots(figsize=(8, 8))
plot_axes(ax, stencil=[(0, 0), (1, 1), (-1, -1), (1, -1), (-1, 1)], coeffs=[-2, 0.5, 0.5, 0.5, 0.5], markersize=40)
fig.savefig("laplace2d-x.png")



import matplotlib.pyplot as plt
import numpy as np
from findiff import FinDiff


def get_started_plot_1():
    x = np.linspace(-np.pi, np.pi, 31)
    dx = x[1] - x[0]
    f = np.sin(x)
    d_dx = FinDiff(0, dx)
    df_dx = d_dx(f)

    plt.xlabel('x')
    plt.ylabel('f, df_dx')
    plt.plot(x, f, '-o', label='f=sin(x)')
    plt.plot(x, df_dx, '-o', label='df_dx' )
    plt.grid()
    plt.legend()

    plt.savefig('get_started_plot_1.png')


if __name__ == '__main__':
    get_started_plot_1()



from sympy import IndexedBase, Symbol, Add, Indexed, latex
from sympy import Rational, factorial, linsolve, Matrix


class SymbolicMesh:
    """
    Represents the mesh on which to evaluate finite difference approximations.
    """

    def __init__(self, coord, equidistant=True):
        """Constructor.

        Parameters
        ----------
        coord: str
            A comma-separated string of coordinate names for the mesh,
            e.g. "x,y" or simply "x"

        equidistant: bool
            Flag indicating whether the mesh is equidistant.
        """
        assert isinstance(coord, str)

        self.equidistant = equidistant
        self._coord_names = [n.replace(" ", "") for n in coord.split(",")]
        self._coord = [IndexedBase(name) for name in self._coord_names]

    @property
    def ndims(self):
        """The number of dimensions of the mesh."""
        return len(self._coord)

    @property
    def coord(self):
        """
        Returns a tuple with the symbols for the coordinates.
        """
        return self._coord

    @property
    def spacing(self):
        """
        Returns a tuple with the spacing of the mesh along all axes.
        Only makes sense for equidistant grid. Raises exception in
        case of non-equidistant grids.
        """
        if self.equidistant:
            spacings = tuple(Symbol(f"\\Delta {x}") for x in self._coord_names)
            return spacings
        raise Exception("Non-equidistant mesh does not have spacing property.")

    @staticmethod
    def create_symbol(name):
        """
        Creates a *sympy* symbol of a given name which can carry as many
        indices as the mesh has dimensions.

        Parameters
        ----------
        name: str
            The name of the meshed symbol.

        Returns
        -------
        An index-carrying *sympy* symbol (IndexedBase).
        """
        return IndexedBase(name)


class SymbolicDiff:
    """
    A symbolic representation of the finite difference approximation
    of a partial derivative. Based on *sympy*.
    """

    def __init__(self, mesh, axis=0, degree=1):
        """Constructor

        Parameters
        ----------
        mesh: SymbolicMesh
            The symbolic grid on which to evaluate the derivative.
        axis: int
            The index of the axis with respect to which to differentiate.
        degree: int > 0
            The degree of the partial derivative.

        """
        self.mesh = mesh
        self.axis = axis
        self.degree = degree

    def __call__(self, f, at, offsets):
        if not isinstance(at, tuple) and not isinstance(at, list):
            at = [at]

        if self.mesh.ndims != len(at):
            raise ValueError("Index tuple must match the number of dimensions!")

        coefs = self._compute_coefficients(f, at, offsets)
        terms = []
        for coef, off in zip(coefs, offsets):
            inds = list(at)
            inds[self.axis] += off
            inds = tuple(inds)
            terms.append(coef * f[inds])

        return Add(*terms).simplify()

    def _compute_coefficients(self, f, at, offsets):

        n = len(offsets)
        # the first row always contains 1s:
        matrix = [[1] * n]

        def spac(off):
            """A helper function to get the spacing between grid points."""
            if self.mesh.equidistant:
                h = self.mesh.spacing[self.axis]
            else:
                x = self.mesh.coord[self.axis]
                h = x[at[self.axis] + off] - x[at[self.axis]]
            return h

        # build up the matrix incrementally:
        for i in range(1, n):
            ifac = Rational(1, factorial(i))
            row = [ifac * (off * spac(off)) ** i for off in offsets]
            matrix.append(row)

        # only the entry corresponding to the requested derivative degree
        # is non-zero:
        rhs = [0] * n
        rhs[self.degree] = 1

        # solve the equation system
        matrix = Matrix(matrix)
        rhs = Matrix(rhs)
        sol = linsolve((matrix, rhs))
        return list(sol)[0].simplify()
